﻿using System;
using System.Reflection;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.IO;
using System.Collections;
using System.Data;

namespace CodeChannelTester
{
    // The main window which holds the main test controller logic
    public partial class MainWindow : Window
    {

        // Complete listing of all found DataSources/Codes/Channels
        // from all the loaded plugins
        private List<DataSource> Available_DataSource;
        private List<Code>       Available_Code;
        private List<Channel>    Available_Channel;    

        // The DataSource/Code/Channel to use based on the users
        // current selection of the given item type
        private DataSource Selected_DataSource;
        private Code       Selected_Code;
        private Channel    Selected_Channel;

        // Tracking state for whether or the not the selected 
        // DataSource/Code/Channel has been set up
        // An item is not set up if it has been unselected or reconfigured.
        // Setup will be performed when testing begins if required.
        private Boolean IsSetup_DataSource;
        private Boolean IsSetup_Code;
        private Boolean IsSetup_Channel;

        // Datatable for logging test results
        private DataTable Sample_Log;


        public MainWindow()
        {
            // Find all possible DataSource/Code/Channel plugins
            RefreshPlugins();

            InitializeComponent();

            // Configure our DataSource/Code/Channel drop-downs to use the found list as options
            cmbDataSource.DisplayMemberPath = "Name";
            cmbDataSource.SetBinding( ComboBox.ItemsSourceProperty, new Binding(){ Source=Available_DataSource });
            cmbCode.DisplayMemberPath       = "Name";
            cmbCode.SetBinding(       ComboBox.ItemsSourceProperty, new Binding(){ Source=Available_Code       });
            cmbChannel.DisplayMemberPath    = "Name";
            cmbChannel.SetBinding(    ComboBox.ItemsSourceProperty, new Binding(){ Source=Available_Channel    });

        }

        // Check the plugins subdirectory to load all plugin .dlls
        // and keep track of the possible DataSources/Codes/Channels found
        public void RefreshPlugins()
        {
            Available_DataSource = new List<DataSource>();
            Available_Code       = new List<Code>();
            Available_Channel    = new List<Channel>();

            var Application_Directory = new Uri(System.Reflection.Assembly.GetExecutingAssembly().CodeBase).LocalPath;
            var Plugins_Directory = Path.Combine(Path.GetDirectoryName(Application_Directory), "Plugins");
            foreach( var dll in Directory.EnumerateFiles(Plugins_Directory, "*.dll", SearchOption.AllDirectories ) )
            {
                try {
                    LoadPlugin( dll );
                }
                catch (Exception ex) {
                    System.Windows.MessageBox.Show( ex.ToString() );
                }
            }

        }

        // Check each class (or, Type, actually) found in a .dll to see if it contains
        // a valid DataSources/Codes/Channels, and track them.
        public void LoadPlugin(String filename)
        {
            var dll = Assembly.LoadFile( filename );
            foreach( Type dllType in dll.GetExportedTypes() )
            {   
                if( typeof(DataSource).IsAssignableFrom(dllType) )
                {   
                    Object obj = Activator.CreateInstance(dllType);
                    Available_DataSource.Add( (DataSource) obj );
                }
                if( typeof(Code).IsAssignableFrom(dllType) )
                {
                    Object obj = Activator.CreateInstance(dllType);
                    Available_Code.Add( (Code) obj );
                }
                if( typeof(Channel).IsAssignableFrom(dllType) )
                {
                    Object obj = Activator.CreateInstance(dllType);
                    Available_Channel.Add( (Channel) obj );
                }
            }
        }

        // A (different) DataSource/Code/Channel has been selected
        private void cmbDataSource_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            Selected_DataSource = (DataSource)cmbDataSource.SelectedValue;
            btnDataSource.IsEnabled = Selected_DataSource.Configurable;
            IsSetup_DataSource = false;
        }
        private void cmbCode_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            Selected_Code = (Code)cmbCode.SelectedValue;
            btnCode.IsEnabled = Selected_Code.Configurable;
            IsSetup_Code = false;
        }
        private void cmbChannel_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            Selected_Channel = (Channel)cmbChannel.SelectedValue;
            btnChannel.IsEnabled = Selected_Channel.Configurable;
            IsSetup_Channel = false;
        }

        // The user is requesting to configure a DataSource/Code/Channel
        private void btnDataSource_Click(object sender, RoutedEventArgs e)
        {
            if(Selected_DataSource==null) return;
            Selected_DataSource.Configure();
            IsSetup_DataSource = false;
        }
        private void btnCode_Click(object sender, RoutedEventArgs e)
        {
            if(Selected_Code==null) return;
            Selected_Code.Configure();
            IsSetup_Code = false;
        }
        private void btnChannel_Click(object sender, RoutedEventArgs e)
        {
            if(Selected_Channel==null) return;
            Selected_Channel.Configure();
            IsSetup_Channel = false;
        }

        // User is requesting to begin testing!
        private void btnGo_Click(object sender, RoutedEventArgs e)
        {
            // Is everthing configured correctly?
            if ( Selected_DataSource==null || Selected_Code==null || Selected_Channel==null ||
                (Selected_DataSource.Configurable && ! Selected_DataSource.Configured) ||
                (Selected_Code.Configurable       && ! Selected_Code.Configured      ) ||
                (Selected_Channel.Configurable    && ! Selected_Channel.Configured   ) )
            {
                StringBuilder warning = new StringBuilder();

                if (Selected_DataSource==null)
                    warning.AppendLine("DataSource not selected.");
                else if (Selected_DataSource.Configurable && ! Selected_DataSource.Configured)
                    warning.AppendLine("DataSource requires configuration.");

                if (Selected_Code==null)
                    warning.AppendLine("Code not selected.");
                else if (Selected_Code.Configurable && ! Selected_Code.Configured)
                    warning.AppendLine("Code requires configuration.");

                if (Selected_Channel==null)
                    warning.AppendLine("Channel not selected.");
                else if (Selected_Channel.Configurable && ! Selected_Channel.Configured)
                    warning.AppendLine("Channel requires configuration.");

                System.Windows.MessageBox.Show( warning.ToString() );
                return;
            }

            // If anything is not set up, set up now.
            if(!IsSetup_DataSource)
            {
                Selected_DataSource.Setup();
                IsSetup_DataSource = true;
            }
            if(!IsSetup_Code)
            {
                Selected_Code.Setup();
                IsSetup_Code = true;
            }
            if(!IsSetup_Channel)
            {
                Selected_Channel.Setup();
                IsSetup_Channel = true;
            }

            // Reset any temporary state.
            Reset();

            // Begin the main testing loop.
            TestLoop();

            // Finalize anything required to save test results
            SaveResults();
        }

        // Reset any temporary state.
        // Includes logging, and any temporary state of the selected DataSource/Code/Channel
        private void Reset()
        {
            // Create a new result log. Intialize it with the required columns.
            Sample_Log = new DataTable();
            var dt = Sample_Log;
            dt.Columns.Add("SampleNumber", typeof(long));      // the index of the sample
            dt.Columns.Add("Input",        typeof(BitArray));  // the raw data of the sample
            dt.Columns.Add("Encoded",      typeof(SoftBurst)); // the encoded value of the sample
            dt.Columns.Add("Error",        typeof(SoftBurst)); // the error introduced by the channel
            dt.Columns.Add("Received",     typeof(SoftBurst)); // the stream received by the decoder
            dt.Columns.Add("Output",       typeof(BitArray));  // the decoded stream
            dt.Columns.Add("Success",      typeof(bool));      // Was the decoded stream the raw stream?

            Selected_DataSource.Reset();
            Selected_Code.Reset();
            Selected_Channel.Reset();
            
        }

        // The main testing loop.
        private void TestLoop()
        {
            int       SampleNum = 0;
            BitArray  burst_raw = null;
            SoftBurst burst_enc = null;
            SoftBurst burst_err = null;
            SoftBurst burst_rcv = null;
            BitArray  burst_dec = null;
            bool      success   = false;
            var       dt        = Sample_Log;

            // continue loop while the datasource produces non-null values.
            for( burst_raw = Selected_DataSource.GetNextBurst();
                 burst_raw != null;
                 burst_raw = Selected_DataSource.GetNextBurst()
            ){


                // From the raw values, produce an encoded stream
                burst_enc = Selected_Code.EncodeBurst((BitArray)burst_raw.Clone());

                // from the encoded stream, pass it through a channel to get a received stream
                burst_rcv = Selected_Channel.CarryTransmission(burst_enc.Copy());

                // determine the error intruded by the channel.
                burst_err = Selected_Channel.LastError;
                // If no explicitly available, determine error by computing difference
                if(Selected_Channel.LastError == null) 
                    burst_err = SoftBurst.EditDifference(burst_enc, burst_rcv);

                // from the received stream, decode it into a decoded stream
                burst_dec = Selected_Code.DecodeBurst(burst_enc.Copy());

                // if the raw and decoded stream are equal, the transmisison succeeded!
                success = BitArrayEqual(burst_raw,burst_dec);


                // Log results of transmitted stream to data table
                var dr = dt.NewRow();
                dr["SampleNumber"] = ++SampleNum;
                dr["Input"]        = burst_raw;
                dr["Encoded"]      = burst_enc;
                dr["Error"]        = burst_err;
                dr["Received"]     = burst_rcv;
                dr["Output"]       = burst_dec;
                dr["Success"]      = success;
                dt.Rows.Add(dr);
                
            }

        }

        public void SaveResults()
        {
            // Now handled by manual exports
        }

        // Determines equality for BitArrays.
        // Perhaps slightly misleading, determines that the first bit array 
        // is the leftmost substring of the second array.
        // Not true equality, but works for purposes of codes padding the end of the signal
        // if input data is not aligned to block coding.
        private bool BitArrayEqual( BitArray b1, BitArray b2 )
        {
            if(b1.Count < b2.Count)
                return false;
            for(int i = 0; i < b1.Count; i++)
                if(b1[i] != b2[i])
                    return false;
            return true;
        }

        // A sligtly more informative string output of a BitArray object.
        // Produces a string of concatenated 1s and 0s, as is common.
        private string BitArrayString( BitArray ba )
        {
            StringBuilder sb = new StringBuilder();
            for(int i = 0; i < ba.Count; i++)
                sb.Append( ba[i] ? "1" : "0" );
            return sb.ToString();
        }

        // A sligtly more informative string output of a SoftBurst object.
        // Produces a space-separated string of double value with fixed width formatting.
        private string SoftBurstString (SoftBurst burst)
        {
            StringBuilder sb = new StringBuilder();
            for(int i = 0; i < burst.BurstData.Length; i++)
            {
                // Clamped to -1000 < x < 1000 with 2 points of decimal precision
                // to ensure that each element is displayed with exactly 7 characters.
                var val = clamp(burst.BurstData[i], -999.99, 999.99);

                // signed. negative values already recieve a sign in the format below.
                var str = (val < 0.0) ? "" : "+";
                str += String.Format( "{0:000.00}", val);

                sb.Append( str );
                if( i < burst.BurstData.Length-1 ) sb.Append(" ");
            }
            return sb.ToString();
        }

        //Constrains a given value to be within a provided maximum and minimum.
        private static double clamp( double val, double min, double max )
        {
            if(val >= max) return max;
            if(val <= min) return min;
            return val;
        }

        // Generate and save result data to a file.
        // Exported data is in the form of a comma-separated value file.
        private void DT_to_CSV( DataTable dt, String filename )
        {
            StringBuilder csv = new StringBuilder();
            
            for(int x = 0; x < dt.Columns.Count; x++)
            {
                csv.Append( "\"" );
                csv.Append( dt.Columns[x].ColumnName );
                csv.Append( "\"" );
                csv.Append( (x < dt.Columns.Count-1) ? ", " : Environment.NewLine );
            }

            for(int y = 0; y < dt.Rows.Count;    y++)
            for(int x = 0; x < dt.Columns.Count; x++)
            {
                csv.Append( "\"" );
                if(dt.Columns[x].DataType == typeof(BitArray))
                    csv.Append(BitArrayString((BitArray)dt.Rows[y][x]));
                else if(dt.Columns[x].DataType == typeof(SoftBurst))
                    csv.Append(SoftBurstString((SoftBurst)dt.Rows[y][x]));
                else
                    csv.Append(dt.Rows[y][x].ToString());
                csv.Append( "\"" );

                csv.Append( (x < dt.Columns.Count-1) ? ", " : Environment.NewLine );
            }

            File.WriteAllText( filename, csv.ToString() );
        }

        // Generate and save result data to a file.
        // Exported data is in the form of each column being printed on a new line.
        // Rows are terminated with an additional blank line.
        private void DT_to_Lines( DataTable dt, String filename )
        {
            StringBuilder lines = new StringBuilder();
            
            for(int x = 0; x < dt.Columns.Count; x++)
                lines.AppendLine( dt.Columns[x].ColumnName );
            lines.AppendLine();

            for(int y = 0; y < dt.Rows.Count;    y++)
            {
                for(int x = 0; x < dt.Columns.Count; x++)
                {   
                    if(dt.Columns[x].DataType == typeof(BitArray))
                        lines.AppendLine(BitArrayString((BitArray)dt.Rows[y][x]));
                    else if(dt.Columns[x].DataType == typeof(SoftBurst))
                        lines.AppendLine(SoftBurstString((SoftBurst)dt.Rows[y][x]));
                    else
                        lines.AppendLine(dt.Rows[y][x].ToString());
                }
                lines.AppendLine();
            }
            File.WriteAllText( filename, lines.ToString() );
        }

        // Export result data is in the form of CSV file.
        private void btnExportCSV_Click(object sender, RoutedEventArgs e)
        {
            // Determine where to save the export.
            if (Sample_Log == null) return;
            try {
                var dialog = new Microsoft.Win32.SaveFileDialog();
                dialog.FileName = "output1";
                dialog.DefaultExt = ".csv";
                dialog.Filter = "Comma-Separated Values File (.csv)|*.csv";
                var dialog_result = dialog.ShowDialog(this);
                if(!(dialog_result??false)) return;

                //Generate and save it.
                DT_to_CSV(Sample_Log, dialog.FileName);                
            }
            catch (Exception ex) {
                System.Windows.MessageBox.Show( ex.ToString() );
            }
        }

        // Export result data in the form of each column being printed on a new line.
        // Rows are terminated with an additional blank line.
        private void btnExportLines_Click(object sender, RoutedEventArgs e)
        {
            // Determine where to save the export.
            if (Sample_Log == null) return;
            try {
                var dialog = new Microsoft.Win32.SaveFileDialog();
                dialog.FileName = "output1";
                dialog.DefaultExt = ".txt";
                dialog.Filter = "Listing of cells on separate lines (.txt)|*.txt";
                var dialog_result = dialog.ShowDialog(this);
                if(!(dialog_result??false)) return;

                //Generate and save it.
                DT_to_Lines(Sample_Log, dialog.FileName);                
            }
            catch (Exception ex) {
                System.Windows.MessageBox.Show( ex.ToString() );
            }
        }

    }
}
