﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CodeChannelTester;
using System.Collections;

namespace Plugin_Channel_SimpleNoise
{
    public class Plugin_Channel_SimpleNoise : Channel
    {

        public override String  Name        { get { return "Simple Noise"; } }
        public override String  Description { get { return "Channel induces a simple tunable additive/multiplicative noise."; } }
        public override Version Version     { get { return new Version(1,0); } }

        public override Boolean Configurable { get { return true; } }
        public override Boolean Configured   { get { return true; } }
        

        // RNG seed. Default is null, resulting in the use of clock time for a seed.
        public int? RNG_Seed = null;
        // RNG Type. Gaussian or Uniform distributions.
        public RNG_Type RNG_Type = RNG_Type.Gaussian;
        // Amplitude is 1.1. An additive noise needs to exceed 1.0 to have a hope of causing decode error
        public double Noise_Amp = 1.1;
        // Mean should usually be 0.0, but we can skew the signal to have more false positives than negatives.
        public double Noise_Mean = 0.0;
        // Variance of the noise distribution. Base variance for the implemented RNG has not been conclusively
        // determined, so there's no need to adjust this for the time being.
        public double Noise_Variance = 1.0;
        // Recombination behaviour between transmission and error signals.
        // Additive noise simply adds the signals together.
        public Noise_Behaviour Noise_Behave = Noise_Behaviour.Additive;

        public RNG RNG = null;


        public override void Configure()
        {
            var dlg = new Configure(this);
            dlg.ShowDialog();
        }

        public override void Reset()
        {
            // Create a new instance of the desired RNG type
            switch( RNG_Type )
            {
                case RNG_Type.Uniform :
                    RNG = new Uniform();
                    break;
                case RNG_Type.Gaussian :
                    RNG = new Gaussian();
                    break;
            }

            // Seed the RNG (and reset it) appropriately.
            RNG.Reset( 
                RNG_Seed ?? DateTime.Now.Millisecond,
                RNG_Seed ?? (int)DateTime.Now.Ticks
            );

        }


        public override SoftBurst LastError { get { return _LastError; } }
        private SoftBurst _LastError = null;


        public override SoftBurst CarryTransmission(SoftBurst transmission)
        {   
            _LastError = new SoftBurst();
            _LastError.BurstData = new double[transmission.BurstData.Length];

            // For each bit of the signal, determine an error. 
            for(int i = 0; i < _LastError.BurstData.Length; i++)
            {
                // Convert the distribution from [0,1], avg=0.5
                // to a distribution on [-Amp,Amp], avg=Mean.
                // Variance can be adjusted, but the variance of the implemented RNG has not yet been discerned.
                var error = Noise_Mean + ( (RNG.GetNext() - 0.5) * Math.Sqrt(Noise_Variance) ) * 2 * Noise_Amp;
                if(Noise_Behave == Noise_Behaviour.Multiplicative)
                    error = Math.Abs(error);

                // Track it and apply it to the transmission signal.
                _LastError.BurstData[i] = error;

                switch(Noise_Behave)
                {
                    case Noise_Behaviour.Additive:
                        transmission.BurstData[i] += _LastError.BurstData[i];
                        break;
                    case Noise_Behaviour.Multiplicative:
                        transmission.BurstData[i] *= _LastError.BurstData[i];
                        break;
                }
            }

            return transmission;
        }

        

    }


    // Enumeration for possible recombination behaviour of transmisison and error signals
    public enum Noise_Behaviour
    {
        // Transmission and noise signals added
        Additive,
        // Transmission signal and absolute value of noise signal multiplied.
        // With the current testing model, many of the deaults will not make good use of multiplicative testing,
        // as a polar encoding is largely unaffected by multiplaction of positive values, where multiplication of
        // signed values often induces far too much variance.
        Multiplicative
    }

}
