﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Collections;

namespace CodeChannelTester
{
    /// <summary>
    /// SoftBurst encapsulates a finite length signal to be transmitted over a channel.
    /// Each bit of data to be transmitted is represented as a sample of the signal.
    /// These samples are held a a double-precision floating point value, and it is
    /// expected that binary data takes a polarized encoding, giving -1.0 and +1.0.
    /// </summary>
    public class SoftBurst
    {
        // The list of signal samples / soft bits.
        public double[] BurstData;
        

        // creates a deep copy of the SoftBurst
        public SoftBurst Copy()
        {
            var sb = new SoftBurst();
            sb.BurstData = new double[BurstData.Length];
            Array.Copy( BurstData, sb.BurstData, sb.BurstData.Length );
            return sb;
        }


        // Static helper method; produces a SoftBurst by applying a polar encoding on 
        // a BitArray
        public static SoftBurst PolarEncode( BitArray bits )
        {
            var sb = new SoftBurst();
            sb.BurstData = new double[ bits.Count ];
            for( int i = 0; i < bits.Count; i++ )
                sb.BurstData[i] = bits[i] ? 1.0 : -1.0 ;
            return sb;
        }

        // Static helper method; produces a bit array by attempting to quantize the
        // signals to -1.0, +1.0 values used in the polar decoding.
        // Simplified by using sign to determine the most likely bit value.
        public static BitArray QuantizeDecode( SoftBurst data ) // A Priori
        {
            var ba = new BitArray( data.BurstData.Length );
            for( int i = 0; i < data.BurstData.Length; i++ )
                ba[i] = data.BurstData[i] >= 0.0;
            return ba;
        }

        // Produces a SofBurst which holds the difference between two other SofBursts.
        // The difference of an element is the signed difference between the first and
        // second SoftBirst arguments. If the two compared values are not of equal length,
        // the difference is found by assuming a value of 0 for each missing signal sample.
        public static SoftBurst EditDifference( SoftBurst b1, SoftBurst b2 )
        {
            var diff = new SoftBurst();
            var length = (int)Math.Max( b1.BurstData.Length, b2.BurstData.Length );
            diff.BurstData = new double[length];

            for(int i = 0; i < length; i++)
            {
                if(i < b1.BurstData.Length && i < b2.BurstData.Length)
                    diff.BurstData[i] =  b2.BurstData[i] - b1.BurstData[i];
                else if (i >= b1.BurstData.Length)
                    diff.BurstData[i] =  b2.BurstData[i];
                else if (i >= b2.BurstData.Length)
                    diff.BurstData[i] = -b1.BurstData[i];
                else
                    break;
            }

            return diff;
        }

    }
}
